/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.processor.processor;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;

import java.time.Duration;
import java.util.Collections;
import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.processor.processor.ProcessorProperties.Processor;
import de.ozgcloud.processor.result.ProcessorTechnicalException;
import de.ozgcloud.processor.result.ResultService;
import de.ozgcloud.processor.vorgang.Vorgang;
import de.ozgcloud.processor.vorgang.VorgangTestFactory;
import lombok.SneakyThrows;
import okhttp3.mockwebserver.MockResponse;
import okhttp3.mockwebserver.MockWebServer;
import okhttp3.mockwebserver.RecordedRequest;
import reactor.core.publisher.Mono;

@ITCase
class ProcessorServiceITCase {

	@Autowired
	private ProcessorService service;
	@SpyBean
	private ResultService resultService;

	private Processor processor;

	@Nested
	class TestCallProcessor {

		private MockWebServer mockWebServer = new MockWebServer();

		@Mock
		private ProcessorResult processorResult;

		@BeforeEach
		void setup() {
			processor = ProcessorTestFactory.createBuilder().address(mockWebServer.url("/").toString()).build();
			mockWebServer.enqueue(createMockResponse());
		}

		private MockResponse createMockResponse() {
			return new MockResponse()
					.addHeader("Content-Type", "application/json")
					.setBody("{\"%s\": \"%s\"}".formatted(ProcessorResultTestFactory.KEY_RESULT_ACTION, ProcessorResultTestFactory.RESULT_ACTION));
		}

		@AfterEach
		@SneakyThrows
		void cleanup() {
			mockWebServer.shutdown();
		}

		@Test
		void shouldSendVorgangIdAsString() {
			var vorgang = VorgangTestFactory.create();

			var request = sendRequest(processor, vorgang);

			assertThat(request.getBody().readUtf8()).contains("\"vorgangId\":\"" + vorgang.getId() + "\"");
		}

		@SneakyThrows
		private RecordedRequest sendRequest(Processor processor, Vorgang vorgang) {
			service.callProcessor(processor, vorgang).block(Duration.ofMillis(1000));
			return mockWebServer.takeRequest(100, TimeUnit.MILLISECONDS);
		}
	}

	@Nested
	class TestProcessorResponse {

		@Mock
		private ProcessorResult processorResult;

		@Nested
		class TestActionField {

			private MockWebServer mockWebServer = new MockWebServer();

			@BeforeEach
			void setup() {
				processor = ProcessorTestFactory.createBuilder().address(mockWebServer.url("/").toString()).build();
			}

			@AfterEach
			@SneakyThrows
			void cleanup() {
				mockWebServer.shutdown();
			}

			@Test
			void shouldAcceptSmallCaseAction() {
				mockWebServer.enqueue(
						createResponse(ProcessorResultTestFactory.KEY_RESULT_ACTION, ProcessorResultTestFactory.RESULT_ACTION.name().toLowerCase()));

				var result = sendRequest(VorgangTestFactory.create());

				assertThat(result.getAction()).isEqualTo(ProcessorResultTestFactory.RESULT_ACTION);
			}

			@DisplayName("should fail when")
			@ParameterizedTest(name = "action is ''{0}''")
			@NullAndEmptySource
			@ValueSource(strings = { "unknown" })
			void shouldFailOnUnknownAction(String action) {
				mockWebServer.enqueue(createResponse(ProcessorResultTestFactory.KEY_RESULT_ACTION, action));
				var vorgang = VorgangTestFactory.create();

				assertThrows(ProcessorTechnicalException.class, () -> sendRequest(vorgang));
			}

			@Test
			void shouldFailWhenNoActionField() {
				mockWebServer.enqueue(createResponse("body", null));
				var vorgang = VorgangTestFactory.create();

				assertThrows(ProcessorTechnicalException.class, () -> sendRequest(vorgang));
			}

			private MockResponse createResponse(String actionKey, String actionValue) {
				return new MockResponse()
						.addHeader("Content-Type", "application/json")
						.setBody("{\"%s\": \"%s\"}".formatted(actionKey, actionValue));
			}
		}

		private ProcessorResult sendRequest(Vorgang vorgang) {
			return service.callProcessor(processor, vorgang).block(Duration.ofMillis(1000));
		}

	}

	@Nested
	class TestErrorResponse {

		private MockWebServer mockWebServer;
		private Vorgang vorgang = VorgangTestFactory.create();

		@BeforeEach
		void setup() {
			mockWebServer = new MockWebServer();
			processor = ProcessorTestFactory.createBuilder().address(mockWebServer.url("/").toString()).build();
			ReflectionTestUtils.setField(service, "properties", ProcessorProperties.builder().processor(processor).build());
		}

		@SneakyThrows
		@AfterEach
		void cleanup() {
			mockWebServer.shutdown();
		}

		@Test
		void shouldHandleRequestError() {
			MockResponse response = new MockResponse().setResponseCode(HttpStatus.BAD_REQUEST.value());
			mockWebServer.enqueue(response);

			var results = service.processVorgang(vorgang);

			assertThrows(ProcessorTechnicalException.class, () -> results.forEach(Mono::block));
		}

		@Test
		void shouldHandleServerError() {
			MockResponse response = new MockResponse().setResponseCode(HttpStatus.BAD_GATEWAY.value());
			mockWebServer.enqueue(response);

			var results = service.processVorgang(vorgang);

			assertThrows(ProcessorTechnicalException.class, () -> results.forEach(Mono::block));
		}

		@Test
		void shouldHandleRedirectStatus() {
			mockWebServer.enqueue(new MockResponse().setResponseCode(301));

			var results = service.processVorgang(vorgang);

			assertThrows(ProcessorTechnicalException.class, () -> results.forEach(Mono::block));
		}

		@Test
		void shouldHandleException() {
			MockResponse response = new MockResponse().addHeader("content-type", MediaType.APPLICATION_JSON).setBody("test");
			mockWebServer.enqueue(response);

			var results = service.processVorgang(vorgang);

			assertThrows(ProcessorTechnicalException.class, () -> results.forEach(Mono::block));
		}
	}

	@DisplayName("Process vorgang")
	@Nested
	class TestProcessVorgang {

		private MockWebServer mockWebServer = new MockWebServer();
		private Vorgang vorgang = VorgangTestFactory.create();
		private Processor processor = ProcessorTestFactory.createBuilder().address(mockWebServer.url("/").toString()).build();
		private ProcessorProperties properties = ProcessorProperties.builder().processor(processor).build();

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(service, "properties", properties);
		}

		@SneakyThrows
		@AfterEach
		void cleanup() {
			mockWebServer.shutdown();
		}

		@Test
		void shouldReturnProcessor() {
			var result = service.processVorgang(vorgang, Collections.singleton(ProcessorTestFactory.PROCESSOR_NAME));

			assertThat(result).isNotEmpty();
		}

		@Test
		void shouldReturnEmptyStream() {
			var result = service.processVorgang(vorgang, Collections.singleton("notMatchingProcessorName"));

			assertThat(result).isEmpty();
		}
	}
}